from gi.repository import Gtk


def show_shortcuts_window(parent_win, *args):
    shortcuts_win = Gtk.Builder.new_from_resource(
        '/org/gabmus/gfeeds/ui/shortcutsWindow.xml'
    ).get_object('shortcuts-gfeeds')
    shortcuts_win.props.section_name = 'shortcuts'
    shortcuts_win.set_transient_for(parent_win)
    shortcuts_win.set_modal(True)
    shortcuts_win.present()
    shortcuts_win.show_all()



from gi.repository import Gtk, GLib, Pango
from gfeeds.confManager import ConfManager
from gfeeds.initials_icon import InitialsIcon
from gfeeds.relative_day_formatter import get_date_format
from gfeeds.sidebar_row_popover import RowPopover


class GFeedsSidebarRow(Gtk.ListBoxRow):
    def __init__(self, feeditem, is_saved=False, **kwargs):
        super().__init__(**kwargs)
        self.is_saved = is_saved
        self.get_style_context().add_class('activatable')
        self.feeditem = feeditem
        self.confman = ConfManager()

        self.builder = Gtk.Builder.new_from_resource(
            '/org/gabmus/gfeeds/ui/sidebar_listbox_row.glade'
        )
        self.container_box = self.builder.get_object('container_box')
        self.title_label = self.builder.get_object('title_label')
        self.title_label.set_text(self.feeditem.title)
        self.confman.connect(
            'gfeeds_full_article_title_changed',
            self.on_full_article_title_changed
        )
        self.on_full_article_title_changed()
        self.origin_label = self.builder.get_object('origin_label')
        self.origin_label.set_text(self.feeditem.parent_feed.title)
        self.confman.connect(
            'gfeeds_full_feed_name_changed',
            self.on_full_feed_name_changed
        )
        self.on_full_feed_name_changed()

        self.icon_container = self.builder.get_object('icon_container')
        self.icon = InitialsIcon(
            self.feeditem.parent_feed.title,
            self.feeditem.parent_feed.favicon_path
        )
        self.icon_container.add(self.icon)

        # Date & time stuff is long
        self.date_label = self.builder.get_object('date_label')
        tz_sec_offset = self.feeditem.pub_date.utcoffset().total_seconds()
        glibtz = GLib.TimeZone(
            (
                '{0}{1}:{2}'.format(
                    '+' if tz_sec_offset >= 0 else '',
                    format(int(tz_sec_offset/3600), '02'),
                    format(int(
                        (tz_sec_offset - (int(tz_sec_offset/3600)*3600))/60
                    ), '02'),
                )
            ) or '+00:00'
        )
        self.datestr = GLib.DateTime(
            glibtz,
            self.feeditem.pub_date.year,
            self.feeditem.pub_date.month,
            self.feeditem.pub_date.day,
            self.feeditem.pub_date.hour,
            self.feeditem.pub_date.minute,
            self.feeditem.pub_date.second
        ).to_local().format(get_date_format(self.feeditem.pub_date))
        self.date_label.set_text(
            self.datestr
        )

        self.popover = RowPopover(self)

        self.add(self.container_box)
        self.set_read()

    def on_full_article_title_changed(self, *args):
        self.title_label.set_ellipsize(
            Pango.EllipsizeMode.NONE if self.confman.conf['full_article_title']
            else Pango.EllipsizeMode.END
        )

    def on_full_feed_name_changed(self, *args):
        self.origin_label.set_ellipsize(
            Pango.EllipsizeMode.NONE if self.confman.conf['full_feed_name']
            else Pango.EllipsizeMode.END
        )

    def set_read(self, read=None):
        if read is not None:
            self.feeditem.set_read(read)
        if self.feeditem.read:
            self.set_dim(True)
        else:
            self.set_dim(False)

    def set_dim(self, state):
        for w in (
                self.title_label,
                self.icon
        ):
            if state:
                w.get_style_context().add_class('dim-label')
            else:
                w.get_style_context().remove_class('dim-label')



from urllib.parse import urlparse


def __build_url(scheme='http', netloc='', path=''):
    return urlparse(
        (scheme + '://') +
        netloc +
        path
    )


def sanitize(base, url):
    parsed_base = urlparse(base)
    parsed_url = urlparse(url)
    if not parsed_url.scheme:
        parsed_url = __build_url(
            parsed_base.scheme,
            parsed_url.netloc,
            parsed_url.path
        )
    if not parsed_url.netloc:
        parsed_url = __build_url(
            parsed_url.scheme,
            parsed_base.netloc,
            parsed_url.path
        )
    return parsed_url.geturl()



from gettext import gettext as _
from os.path import isfile
import requests
from lxml.html import html5parser
from gfeeds.confManager import ConfManager
from gfeeds.sha import shasum
from gfeeds.url_sanitizer import sanitize

confman = ConfManager()

GET_HEADERS = {
    'User-Agent': 'gfeeds/1.0',
    'Accept': '*/*',
    'Accept-Encoding': 'gzip, deflate'
}

TIMEOUT = 30


# will return the content of a file if it's a file url
def download_text(link: str) -> str:
    if link[:8] == 'file:///':
        with open(link[7:]) as fd:
            toret = fd.read()
        return toret
    res = requests.get(link, headers=GET_HEADERS, timeout=TIMEOUT)
    if 200 <= res.status_code <= 299:
        res.encoding = 'utf-8'
        return res.text
    else:
        print(f'response code {res.status_code}')
        raise requests.HTTPError(f'response code {res.status_code}')


def download_raw(link: str, dest: str):
    res = requests.get(link, headers=GET_HEADERS, timeout=TIMEOUT)
    if res.status_code == 200:
        with open(dest, 'wb') as fd:
            for chunk in res.iter_content(1024):
                fd.write(chunk)
    else:
        raise requests.HTTPError(
            f'response code {res.status_code} for url `{link}`'
        )


def extract_feed_url_from_html(link: str) -> str:
    try:
        html = download_text(link)
        root = html5parser.fromstring(
            html if type(html) == str else html.decode()
        )
        link_els = root.xpath(
            '//x:link',
            namespaces={'x': 'http://www.w3.org/1999/xhtml'}
        )
        for el in link_els:
            if (
                el.attrib.get('rel', '') == 'alternate' and
                el.attrib.get('type', '') in (
                    'application/atom+xml', 'application/rss+xml'
                ) and 'href' in el.attrib.keys()
            ):
                return sanitize(link, el.attrib['href'])
    except Exception:
        print('exception in `extract_feed_from_html`')
    return None


# TODO: refactor this, the returned typeS(!) are a mess
def download_feed(link: str, get_cached: bool = False):
    dest_path = confman.cache_path.joinpath(shasum(link)+'.rss')
    if get_cached:
        return (dest_path, link) if isfile(dest_path) else ('not_cached', None)
    headers = GET_HEADERS.copy()
    if (
            'last-modified' in confman.conf['feeds'][link].keys() and
            isfile(dest_path)
    ):
        headers['If-Modified-Since'] = \
            confman.conf['feeds'][link]['last-modified']
    try:
        res = requests.get(
            link, headers=headers, allow_redirects=True, timeout=TIMEOUT
        )
    except requests.exceptions.ConnectTimeout:
        return (False, _('`{0}`: connection timed out').format(link))
    except Exception:
        return (False, _('`{0}` is not an URL').format(link))
    if 'last-modified' in res.headers.keys():
        confman.conf['feeds'][link]['last-modified'] = \
            res.headers['last-modified']

    def handle_200():
        if (
                'last-modified' not in res.headers.keys() and
                'last-modified' in confman.conf['feeds'][link].keys()
        ):
            confman.conf['feeds'][link].pop('last-modified')
        with open(dest_path, 'w') as fd:
            fd.write(res.text)
        return (dest_path, link)

    def handle_304(): return (dest_path, link)

    def handle_301_302():
        n_link = res.headers.get('location', link)
        confman.conf['feeds'][n_link] = confman.conf['feeds'][link]
        confman.conf['feeds'].pop(link)
        return download_feed(n_link)

    def handle_everything_else(): return (
        False,
        _('Error downloading `{0}`, code `{1}`').format(link, res.status_code)
    )

    handlers = {
        200: handle_200, 304: handle_304,
        301: handle_301_302, 302: handle_301_302
    }
    return handlers.get(res.status_code, handle_everything_else)()



from gi.repository import Gtk


class ScrolledMessageDialog(Gtk.MessageDialog):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.builder = Gtk.Builder.new_from_resource(
            '/org/gabmus/gfeeds/ui/scrolled_label.glade'
        )
        self.scrolled_win = self.builder.get_object('scrolled_win')
        self.label = self.builder.get_object('label')

        self.get_content_area().pack_start(self.scrolled_win, True, True, 0)
        self.get_content_area().set_spacing(0)
        self.get_message_area().set_spacing(0)
        self.title_label = self.get_message_area().get_children()[0]
        self.secondary_label = self.get_message_area().get_children()[1]
        self.title_label.set_margin_bottom(20)
        # Hiding default secondary text, because I need it set to make the
        # title pretty but I use a custom widget for the secondary text
        self.secondary_label.set_visible(False)
        self.secondary_label.set_no_show_all(True)

    def format_secondary_markup(self, message):
        super().format_secondary_markup(' ')
        self.label.set_markup(message)



from gettext import gettext as _
from gi.repository import Gtk


class RefreshSpinnerButton(Gtk.Bin):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.builder = Gtk.Builder.new_from_resource(
            '/org/gabmus/gfeeds/ui/spinner_button.glade'
        )
        self.btn = self.builder.get_object('spinnerbutton')
        self.stack = self.builder.get_object('stack')
        self.spinner = self.builder.get_object('spinner')
        self.icon = self.builder.get_object('icon')

        self.btn.set_tooltip_text(_('Refresh feeds'))
        self.add(self.btn)
        self.stack.set_visible_child(self.icon)

    def set_spinning(self, state):
        if state:
            self.stack.set_visible_child(self.spinner)
            self.btn.set_sensitive(False)
        else:
            self.stack.set_visible_child(self.icon)
            self.btn.set_sensitive(True)



from gi.repository import Gtk, Handy, GdkPixbuf
from gfeeds.confManager import ConfManager
from pathlib import Path
from PIL import Image


confman = ConfManager()
thumbs_cache_path = Path(confman.thumbs_cache_path)


def make_thumb(path, width: int, height: int = 1000) -> str:
    if not path:
        return None
    if not isinstance(path, Path):
        path = Path(path)
    dest = thumbs_cache_path.joinpath(f'{width}x{height}_{path.name}_v2')
    if dest.is_file():
        return str(dest)
    try:
        with Image.open(path) as thumb:
            thumb = Image.open(path)
            thumb.resize((width, height)).save(dest, 'PNG')
            # thumb.thumbnail((width, height), Image.ANTIALIAS)
            # thumb.save(dest, 'PNG')
        return str(dest)
    except IOError:
        print(f'Error creating thumbnail for image `{path}`')
        return None


avatar_cache = dict()


def set_avatar_func(icon: str, size: int) -> GdkPixbuf.Pixbuf:
    if icon is None:
        return None
    key = f'{icon}__{size}'
    if key in avatar_cache.keys():
        return avatar_cache[key]
    pixbuf = None
    try:
        pixbuf = GdkPixbuf.Pixbuf.new_from_file(
            make_thumb(icon, size, size)
        )
    except Exception:
        print(f'Error creating pixbuf for icon `{icon}`')
    avatar_cache[key] = pixbuf
    return pixbuf


class InitialsIcon(Gtk.Bin):
    def __init__(self, name, image_path, **kwargs):
        super().__init__(**kwargs)
        self.name = name
        self.image_path = image_path
        self.avatar = Handy.Avatar.new(
            32,
            self.name,
            True
        )
        self.avatar.set_image_load_func(
            lambda size: set_avatar_func(self.image_path, size)
        )
        self.add(self.avatar)



from gi.repository import Gtk
from gfeeds.feeds_manager import FeedsManager


class StackWithEmptyState(Gtk.Stack):
    def __init__(self, main_widget, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.feedman = FeedsManager()
        self.main_widget = main_widget
        self.empty_state = Gtk.Builder.new_from_resource(
            '/org/gabmus/gfeeds/ui/empty_state.glade'
        ).get_object('empty_state_box')
        self.main_widget.show_all()
        self.add_named(self.main_widget, 'main_widget')
        self.add_named(self.empty_state, 'empty_state')
        self.set_visible_child(self.main_widget)

        self.feedman.feeds.connect(
            'pop',
            self.on_feeds_pop
        )
        self.feedman.connect(
            'feedmanager_refresh_end',
            self.on_feeds_pop
        )
        self.feedman.feeds.connect(
            'append',
            self.on_feeds_append
        )

        self.set_transition_type(
            Gtk.StackTransitionType.CROSSFADE
        )

        # self.set_size_request(360, 100)

    def on_feeds_pop(self, *args):
        if len(self.feedman.feeds) == 0:
            self.set_visible_child(self.empty_state)
        else:
            self.set_visible_child(self.main_widget)

    def on_feeds_append(self, *args):
        self.set_visible_child(self.main_widget)



from gi.repository import GObject


class SignalerList(GObject.Object):
    __gsignals__ = {
        'append': (
            GObject.SignalFlags.RUN_LAST,
            None,
            (GObject.TYPE_PYOBJECT,)
        ),
        'extend': (
            GObject.SignalFlags.RUN_LAST,
            None,
            (GObject.TYPE_PYOBJECT,)
        ),
        'pop': (
            GObject.SignalFlags.RUN_LAST,
            None,
            (GObject.TYPE_PYOBJECT,)
        ),
        'empty': (
            GObject.SignalFlags.RUN_LAST,
            None,
            (str,)
        )
    }

    def __init__(self, n_list=None, **kwargs):
        if n_list is None:
            n_list = []
        super().__init__(**kwargs)
        self.__list = n_list.copy()

    # Note: you see all those "item" names?
    # Don't be confused, they don't refer to FeedItem objects
    # It's item as in "list item"

    def index(self, item):
        return self.__list.index(item)

    def extend(self, ex_list):
        self.__list.extend(ex_list)
        self.emit('extend', ex_list)

    def empty(self):
        self.__list = []
        self.emit('empty', '')

    def append(self, n_item):
        self.__list.append(n_item)
        self.emit('append', n_item)

    def pop(self, item):
        popped = self.__list[item]
        self.__list.pop(item)
        self.emit('pop', popped)
        return popped

    def remove(self, item):
        self.__list.remove(item)
        self.emit('pop', item)
        return item

    def __len__(self):
        return len(self.__list)

    # this also provides iteration and slicing
    def __getitem__(self, item):
        return self.__list[item]

    def __repr__(self):
        return str(type(self)) + self.__list.__repr__()

    def get_list(self):
        return self.__list.copy()



from gettext import gettext as _
from gi.repository import Gtk, Pango
from gfeeds.confManager import ConfManager
from gfeeds.feeds_manager import FeedsManager
from gfeeds.initials_icon import InitialsIcon
from gfeeds.listbox_tools import separator_header_func


class FeedsViewAllListboxRow(Gtk.ListBoxRow):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.IS_ALL = True
        self.IS_TAG = False
        self.feed = None
        self.label = Gtk.Label()
        self.label.set_markup(
            '<b>' +
            _('All feeds') +
            '</b>'
        )
        self.label.set_use_markup(True)
        self.label.set_margin_top(12)
        self.label.set_margin_bottom(12)
        self.add(self.label)


class FeedsViewListboxRow(Gtk.ListBoxRow):
    def __init__(self, feed, description=True, **kwargs):
        super().__init__(**kwargs)
        self.confman = ConfManager()
        self.IS_ALL = False
        self.IS_TAG = False
        self.feed = feed
        self.title = feed.title
        self.builder = Gtk.Builder.new_from_resource(
            '/org/gabmus/gfeeds/ui/manage_feeds_listbox_row.glade'
        )
        self.hbox = self.builder.get_object('hbox')
        self.checkbox = self.builder.get_object('check')
        self.checkbox.set_no_show_all(True)
        self.checkbox.hide()

        self.icon_container = self.builder.get_object('icon_container')
        self.icon = InitialsIcon(
            self.feed.title,
            self.feed.favicon_path
        )
        self.icon_container.add(self.icon)

        self.name_label = self.builder.get_object('title_label')
        self.name_label.set_text(self.feed.title)
        self.confman.connect(
            'gfeeds_full_feed_name_changed',
            self.on_full_feed_name_changed
        )
        self.desc_label = self.builder.get_object('description_label')
        self.desc_label.set_no_show_all(not description)
        if description:
            self.desc_label.set_text(self.feed.description)
        else:
            self.desc_label.hide()
            self.name_label.set_ellipsize(Pango.EllipsizeMode.END)
        self.add(self.hbox)
        self.on_full_feed_name_changed()

    def on_full_feed_name_changed(self, *args):
        self.name_label.set_ellipsize(
            Pango.EllipsizeMode.NONE if self.confman.conf['full_feed_name']
            else Pango.EllipsizeMode.END
        )

    def __repr__(self):
        return f'<FeedsViewListboxRow - {self.title}>'


class FeedsViewTagListboxRow(Gtk.ListBoxRow):
    def __init__(self, tag, **kwargs):
        super().__init__(**kwargs)
        self.IS_TAG = True
        self.IS_ALL = False
        self.tag = tag
        self.title = tag
        self.feed = None
        self.builder = Gtk.Builder.new_from_resource(
            '/org/gabmus/gfeeds/ui/manage_feeds_listbox_row.glade'
        )
        self.hbox = self.builder.get_object('hbox')
        self.checkbox = self.builder.get_object('check')
        self.checkbox.set_no_show_all(True)
        self.checkbox.hide()
        self.icon_container = self.builder.get_object('icon_container')
        self.icon = Gtk.Image.new_from_icon_name(
            'tag-symbolic',
            Gtk.IconSize.INVALID
        )
        self.icon.set_pixel_size(32)
        self.icon_container.add(self.icon)

        self.name_label = self.builder.get_object('title_label')
        self.name_label.set_text(tag)
        self.desc_label = self.builder.get_object('description_label')
        self.desc_label.set_no_show_all(True)
        self.desc_label.hide()
        self.add(self.hbox)

    def __repr__(self):
        return f'<FeedsViewTagListboxRow - {self.title}>'


class FeedsViewListbox(Gtk.ListBox):
    def __init__(self, description=True, tags=False, **kwargs):
        super().__init__(**kwargs)
        self.description = description
        self.feedman = FeedsManager()
        self.confman = ConfManager()

        self.set_selection_mode(Gtk.SelectionMode.SINGLE)
        self.connect('row-activated', self.on_row_activated)

        for feed in self.feedman.feeds:
            self.add_feed(feed)
        self.feedman.feeds.connect(
            'empty',
            self.empty
        )
        self.feedman.feeds.connect(
            'append',
            self.on_feeds_append
        )
        self.feedman.feeds.connect(
            'pop',
            self.on_feeds_pop
        )

        if tags:
            self.confman.connect(
                'gfeeds_tags_append',
                self.on_tags_append
            )
            self.confman.connect(
                'gfeeds_tags_pop',
                self.on_tags_pop
            )

            for tag in self.confman.conf['tags']:
                self.on_tags_append(None, tag)

        self.set_sort_func(self.gfeeds_sort_func, None, False)
        self.set_header_func(separator_header_func)

    def on_feeds_pop(self, caller, feed):
        self.remove_feed(feed)

    def on_feeds_append(self, caller, feed):
        self.add_feed(feed)

    def on_tags_append(self, caller, tag):
        self.add(FeedsViewTagListboxRow(tag))

    def on_tags_pop(self, caller, tag):
        for row in self.get_children():
            if row.IS_TAG and row.tag == tag:
                self.remove(row)
                break

    def add_feed(self, feed):
        self.add(FeedsViewListboxRow(feed, self.description))

    def add(self, *args, **kwargs):
        super().add(*args, **kwargs)
        self.show_all()

    def on_row_activated(self, listbox, row):
        if row.IS_ALL:
            self.confman.emit('gfeeds_filter_changed', None)
            return
        if row.IS_TAG:
            self.confman.emit('gfeeds_filter_changed', [row.tag])
            return
        self.confman.emit('gfeeds_filter_changed', row.feed)

    def remove_feed(self, feed):
        for row in self.get_children():
            if not row.IS_ALL:
                if row.feed == feed:
                    self.remove(row)
                    break

    def empty(self, *args):
        rows = self.get_children()
        for row in rows:
            if row and not row.IS_ALL and not row.IS_TAG:
                self.remove(row)

    def row_all_activate(self, skip=False):
        if skip:
            return
        for row in self.get_children():
            if row.IS_ALL:
                row.activate()

    def gfeeds_sort_func(self, row1, row2, data, notify_destroy):
        if row1.IS_ALL:
            return False
        if row2.IS_ALL:
            return True
        if row1.IS_TAG and not row2.IS_TAG:
            return False
        if row2.IS_TAG and not row1.IS_TAG:
            return True
        return row1.title.lower() > row2.title.lower()


class FeedsViewScrolledWindow(Gtk.ScrolledWindow):
    def __init__(self, description=True, tags=False, **kwargs):
        super().__init__(**kwargs)
        self.listbox = FeedsViewListbox(description, tags)
        self.all_row = FeedsViewAllListboxRow()
        self.listbox.add(self.all_row)
        self.listbox.select_row(self.all_row)
        # self.set_size_request(360, 500)
        self.set_policy(Gtk.PolicyType.NEVER, Gtk.PolicyType.AUTOMATIC)
        self.add(self.listbox)
        self.set_size_request(250, 400)
        self.get_style_context().add_class('frame')
        self.set_margin_top(6)
        self.set_margin_end(6)
        self.set_margin_bottom(6)
        self.set_margin_start(6)


class FeedsViewPopover(Gtk.Popover):
    def __init__(self, relative_to, **kwargs):
        super().__init__(**kwargs)
        self.scrolled_win = FeedsViewScrolledWindow(
            description=False,
            tags=True
        )
        self.add(self.scrolled_win)
        self.set_modal(True)
        self.set_relative_to(relative_to)
        relative_to.connect('clicked', self.on_relative_to_clicked)

    def on_relative_to_clicked(self, *args):
        self.popup()
        self.show_all()



import json
import feedparser
import pytz
from datetime import datetime, timezone
from dateutil.parser import parse as dateparse
from dateutil.tz import gettz
from gettext import gettext as _
from gfeeds.download_manager import download_raw
from gfeeds.get_favicon import get_favicon
from os.path import isfile
from gfeeds.confManager import ConfManager
from gfeeds.sha import shasum
from PIL import Image
from bs4 import UnicodeDammit
from typing import Optional


def get_encoding(in_str):
    sample = in_str[:200]
    if 'encoding' in sample:
        enc_i = sample.index('encoding')
        trim = sample[enc_i+10:]
        str_delimiter = "'" if "'" in trim else '"'
        encoding = trim[:trim.index(str_delimiter)]
        return encoding
    return 'utf-8'


class FeedItem:
    def __init__(self, fp_item, parent_feed):
        self.confman = ConfManager()
        self.fp_item = fp_item
        self.is_saved = 'linkhash' in self.fp_item.keys()
        self.title = self.fp_item.get('title', '')
        self.link = self.fp_item.get('link', '')
        self.read = self.link in self.confman.read_feeds_items
        # self.description = self.fp_item.get('description', '')
        self.pub_date_str = self.fp_item.get(
            'published',
            self.fp_item.get('updated', '')
        )
        self.pub_date = datetime.now(timezone.utc)  # fallback to avoid errors
        self.parent_feed = parent_feed

        try:
            self.pub_date = dateparse(self.pub_date_str, tzinfos={
                'UT': gettz('GMT'),
                'EST': -18000,
                'EDT': -14400,
                'CST': -21600,
                'CDT': -18000,
                'MST': -25200,
                'MDT': -21600,
                'PST': -28800,
                'PDT': -25200
            })
            if not self.pub_date.tzinfo:
                self.pub_date = pytz.UTC.localize(self.pub_date)
        except Exception:
            print(_(
                'Error: unable to parse datetime {0} for feeditem {1}'
            ).format(self.pub_date_str, self))

    def set_read(self, read):
        if read == self.read:  # how could this happen?
            return
        if read and self.link not in self.confman.read_feeds_items:
            self.confman.read_feeds_items.append(self.link)
        elif self.link in self.confman.read_feeds_items:
            self.confman.read_feeds_items.remove(self.link)
        self.read = read

    def __repr__(self):
        return 'FeedItem Object `{0}` from Feed {1}'.format(
            self.title,
            self.parent_feed.title
        )

    def to_dict(self):
        return {
            'title': self.title,
            'link': self.link,
            'linkhash': shasum(self.link),
            'published': str(self.pub_date),
            'parent_feed': {
                'title': self.parent_feed.title,
                'link': self.parent_feed.link,
                'favicon_path': self.parent_feed.favicon_path
            }
        }

    def to_json(self):
        return json.dumps(self.to_dict())

    @classmethod
    def new_from_dict(cls, n_fi_dict):
        return cls(
            n_fi_dict,
            FakeFeed(n_fi_dict['parent_feed'])
        )

    @classmethod
    def new_from_json(cls, fi_json):
        return cls.new_from_dict(json.loads(fi_json))


class FakeFeed:
    def __init__(self, f_dict):
        self.title = f_dict.get('title', '')
        self.link = f_dict.get('link', '')
        self.favicon_path = f_dict.get('favicon_path', '')
        if isfile(self.favicon_path):
            favicon = Image.open(self.favicon_path)
            if favicon.width != 32:
                favicon = favicon.resize((32, 32), Image.BILINEAR)
                favicon.save(self.favicon_path, 'PNG')
            favicon.close()

    def __repr__(self):
        return f'FakeFeed Object `{self.title}`'


def parse_feed(feed_str: str, no_preprocessing=False) -> Optional[str]:
    if no_preprocessing:
        return feedparser.parse(feed_str)
    udammit = UnicodeDammit(feed_str)
    feed_str = udammit.unicode_markup
    feed_str = feed_str.replace(
        get_encoding(feed_str),
        'utf-8'
    )
    forbidden_namespaces = [
        'atom',
        'openSearch',
        'thr'
    ]
    for fns in forbidden_namespaces:
        feed_str = feed_str.replace(
            f'<{fns}:', '<'
        )
        feed_str = feed_str.replace(
            f'</{fns}:', '</'
        )
    try:
        fp_feed = feedparser.parse(feed_str)
        return fp_feed
    except Exception:
        import traceback
        traceback.print_exc()
        return None


class Feed:
    def __init__(self, download_res, no_preprocessing=False):
        self.is_null = False
        self.error = None
        if download_res[0] is False:  # indicates failed download
            self.is_null = True
            self.error = download_res[1]
            print(download_res)
            return
        feedpath = download_res[0]
        self.rss_link = download_res[1]
        with open(feedpath, 'r') as fd:
            feed_str = fd.read()
        self.fp_feed = parse_feed(feed_str, no_preprocessing)
        if self.fp_feed is None:
            self.is_null = True
            self.error = _('Errors while parsing feed `{0}`').format(
                self.rss_link
            )
            return

        self.confman = ConfManager()
        self.init_time = pytz.UTC.localize(datetime.utcnow())

        self.fp_feed['feed'] = self.fp_feed.get(
            'feed',
            self.fp_feed.get('channel', {})
        )
        self.title = self.fp_feed.feed.get('title', '')
        self.link = self.fp_feed.feed.get('link', '')
        self.description = self.fp_feed.feed.get('subtitle', self.link)
        # self.language = self.fp_feed.get('', '')
        image_tag = self.fp_feed.get('feed', {}).get('image', {})
        self.image_url = image_tag.get(
            'href',
            image_tag.get('url', '')
        )
        self.items = []
        raw_entries = self.fp_feed.get('entries', [])
        for entry in raw_entries:
            n_item = FeedItem(entry, self)
            item_age = self.init_time - n_item.pub_date
            if item_age < self.confman.max_article_age:
                self.items.append(n_item)
            elif n_item.read:
                self.confman.read_feeds_items.remove(n_item.link)
        self.color = [0.0, 0.0, 0.0]

        if (
                not self.title and
                len(raw_entries) == 0
        ):
            # if these conditions are met, there's reason to believe
            # this is not an rss/atom feed
            self.is_null = True
            self.error = _(
                '`{0}` may not be an RSS or Atom feed'
            ).format(self.rss_link)
            if not no_preprocessing:
                self.__init__(download_res, no_preprocessing=True)
            return

        if not self.title:
            self.title = self.link
            if not self.title:
                self.title = self.rss_link

        self.favicon_path = self.confman.thumbs_cache_path+'/' + \
            shasum(self.link)+'.png'
        if not isfile(self.favicon_path):
            if self.image_url:
                try:
                    download_raw(self.image_url, self.favicon_path)
                except Exception:
                    self.image_url = None
            if not self.image_url:
                try:
                    get_favicon(self.link, self.favicon_path)
                    if not isfile(self.favicon_path):
                        get_favicon(self.items[0].link, self.favicon_path)
                except Exception:
                    print('No favicon')

    def __repr__(self):
        return f'Feed Object `{self.title}`; {len(self.items)} items'



from gi.repository import GLib
from gettext import gettext as _
import listparser
from threading import Thread
from os.path import isfile
from xml.sax.saxutils import escape
from gfeeds.confManager import ConfManager
from gfeeds.feeds_manager import FeedsManager

confman = ConfManager()
feedman = FeedsManager()


def __add_feeds_from_opml_callback(n_feeds_urls_l):
    for tag in [t for f in n_feeds_urls_l for t in f['tags']]:
        confman.add_tag(tag)
    for f in n_feeds_urls_l:
        url = f['feed']
        if url not in confman.conf['feeds'].keys():
            confman.conf['feeds'][url] = {
                'tags': f['tags']
            }
    confman.save_conf()
    feedman.refresh()


def __add_feeds_from_opml_worker(opml_path):
    n_feeds_urls_l = opml_to_rss_list(opml_path)
    GLib.idle_add(
        __add_feeds_from_opml_callback,
        n_feeds_urls_l
    )


def add_feeds_from_opml(opml_path):
    t = Thread(target=__add_feeds_from_opml_worker, args=(opml_path,))
    t.start()


def opml_to_rss_list(opml_path):
    if not isfile(opml_path):
        print(_('Error: OPML path provided does not exist'))
        return []
    try:
        lp_opml = listparser.parse(opml_path)
        n_feeds_urls_l = [
            {'feed': f['url'], 'tags': f['tags']} for f in lp_opml['feeds']
        ]
        return n_feeds_urls_l
    except Exception:
        print(_('Error parsing OPML file `{0}`').format(opml_path))
        return []


OPML_PREFIX = '''<?xml version="1.0" encoding="UTF-8"?>
<opml version="2.0">
  <head>
    <title>Subscriptions</title>
  </head>
  <body>
'''

OPML_SUFFIX = '''
  </body>
</opml>
'''


def feeds_list_to_opml(feeds):
    opml_out = OPML_PREFIX
    for f in feeds:
        opml_out += f'''
            <outline
                title="{escape(f.title)}"
                text="{escape(f.description)}"
                type="rss"
                xmlUrl="{escape(f.rss_link)}"
                htmlUrl="{escape(f.link)}"
                category="{escape(
                    ','.join(confman.conf['feeds'][f.rss_link]['tags'])
                ) if 'tags' in confman.conf['feeds'][f.rss_link].keys()
                else ''}"
            />
        '''
    opml_out += OPML_SUFFIX
    return opml_out



from gfeeds.confManager import ConfManager
from gi.repository import Gio

confman = ConfManager()

sans_font = 'Cantarell'  # confman.sans_font
serif_font = 'DejaVu Serif'  # confman.serif_font
mono_font = 'DejaVu Sans Mono'  # confman.mono_font

CSS = f'''@font-face {{
  font-family: ephy-reader-serif;
  src: local('{serif_font}');
  font-weight: normal;
  font-style: normal;
  font-display: block;
}}

@font-face {{
  font-family: ephy-reader-serif;
  src: local('{serif_font} Italic');
  font-weight: normal;
  font-style: italic;
  font-display: block;
}}

@font-face {{
  font-family: ephy-reader-serif;
  src:local('{serif_font} Bold');
  font-weight: bold;
  font-style: normal;
  font-display: block;
}}

@font-face {{
  font-family: ephy-reader-serif;
  src:local('{serif_font} Bold Italic');
  font-weight: bold;
  font-style: italic;
  font-display: block;
}}

@font-face {{
  font-family: ephy-reader-sans;
  src: local('{sans_font}');
  font-weight: normal;
  font-style: normal;
  font-display: block;
}}

@font-face {{
  font-family: ephy-reader-sans;
  src: local('{sans_font} Italic');
  font-weight: normal;
  font-style: italic;
  font-display: block;
}}

@font-face {{
  font-family: ephy-reader-sans;
  src: local('{sans_font} Bold');
  font-weight: bold;
  font-style: normal;
  font-display: block;
}}

@font-face {{
  font-family: ephy-reader-sans;
  src: local('{sans_font} Bold Italic');
  font-weight: bold;
  font-style: italic;
  font-display: block;
}}

@font-face {{
  font-family: ephy-reader-mono;
  src: local('{mono_font}');
  font-weight: normal;
  font-style: normal;
  font-display: block;
}}

@font-face {{
  font-family: ephy-reader-mono;
  src: local('{mono_font} Italic');
  font-weight: normal;
  font-style: italic;
  font-display: block;
}}

@font-face {{
  font-family: ephy-reader-mono;
  src: local('{mono_font} Bold');
  font-weight: bold;
  font-style: normal;
  font-display: block;
}}

@font-face {{
  font-family: ephy-reader-mono;
  src: local('{mono_font} Bold Italic');
  font-weight: bold;
  font-style: italic;
  font-display: block;
}}
''' + Gio.resources_lookup_data(
    '/org/gabmus/gfeeds/reader_mode_style.css',
    Gio.ResourceLookupFlags.NONE
).get_data().decode()
DARK_MODE_CSS = Gio.resources_lookup_data(
    '/org/gabmus/gfeeds/reader_mode_dark_style.css',
    Gio.ResourceLookupFlags.NONE
).get_data().decode()



from os import remove, listdir
from os.path import isfile, abspath, join
from gettext import gettext as _
from gi.repository import Gtk, Handy
from gfeeds.confManager import ConfManager


def show_settings_window(parent_win, *args):
    settings_win = GFeedsSettingsWindow()
    settings_win.set_transient_for(parent_win)
    settings_win.set_modal(True)
    settings_win.present()


class PreferencesButtonRow(Handy.ActionRow):
    """
    A preferences row with a title and a button
    title: the title shown
    button_label: a label to show inside the button
    onclick: the function that will be called when the button is pressed
    button_style_class: the style class of the button.
        Common options: `suggested-action`, `destructive-action`
    signal: an optional signal to let ConfManager emit when the button is
        pressed
    """
    def __init__(self, title, button_label, onclick, button_style_class=None,
                 signal=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.title = title
        self.button_label = button_label
        self.confman = ConfManager()
        self.set_title(self.title)
        self.signal = signal
        self.onclick = onclick

        self.button = Gtk.Button()
        self.button.set_label(self.button_label)
        self.button.set_valign(Gtk.Align.CENTER)
        if button_style_class:
            self.button.get_style_context().add_class(button_style_class)
        self.button.connect('clicked', self.on_button_clicked)
        self.add(self.button)
        # You need to press the actual button
        # Avoids accidental presses
        # self.set_activatable_widget(self.button)

    def on_button_clicked(self, button):
        self.onclick(self.confman)
        if self.signal:
            self.confman.emit(self.signal, '')
        self.confman.save_conf()


class PreferencesSpinButtonRow(Handy.ActionRow):
    """
    A preferences row with a title and a spin button
    title: the title shown
    min_v: minimum num value
    max_v: maximum num value
    conf_key: the key of the configuration dictionary/json in ConfManager
    signal: an optional signal to let ConfManager emit when the value changes
    """

    def __init__(self, title, min_v, max_v, conf_key, signal=None,
                 subtitle=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.title = title
        if subtitle:
            self.subtitle = subtitle
            self.set_subtitle(self.subtitle)
        self.confman = ConfManager()
        self.set_title(self.title)
        self.signal = signal
        self.conf_key = conf_key

        self.adjustment = Gtk.Adjustment(
            self.confman.conf[self.conf_key],  # initial value
            min_v,  # minimum value
            max_v,  # maximum value
            1,  # step increment
            7,  # page increment (page up, page down? large steps anyway)
            0
        )

        self.spin_button = Gtk.SpinButton()
        self.spin_button.set_adjustment(self.adjustment)
        self.spin_button.set_valign(Gtk.Align.CENTER)
        self.spin_button.connect('value-changed', self.on_value_changed)
        self.add(self.spin_button)
        # You need to interact with the actual spin button
        # Avoids accidental presses
        # self.set_activatable_widget(self.button)

    def on_value_changed(self, *args):
        self.confman.conf[self.conf_key] = self.spin_button.get_value_as_int()
        if self.signal:
            self.confman.emit(self.signal, '')
        self.confman.save_conf()


class PreferencesToggleRow(Handy.ActionRow):
    """
    A preferences row with a title and a toggle
    title: the title shown
    conf_key: the key of the configuration dictionary/json in ConfManager
    signal: an optional signal to let ConfManager emit when the configuration
        is set
    """
    def __init__(self, title, conf_key, signal=None, subtitle=None,
                 *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.title = title
        self.subtitle = subtitle
        if self.subtitle is not None:
            self.set_subtitle(self.subtitle)
        self.confman = ConfManager()
        self.set_title(self.title)
        self.conf_key = conf_key
        self.signal = signal

        self.toggle = Gtk.Switch()
        self.toggle.set_valign(Gtk.Align.CENTER)
        if self.conf_key == 'selection_mode':
            self.toggle.set_active(
                self.confman.conf[self.conf_key] == 'double'
            )
        else:
            self.toggle.set_active(self.confman.conf[self.conf_key])
        self.toggle.connect('state-set', self.on_toggle_state_set)
        self.add(self.toggle)
        self.set_activatable_widget(self.toggle)

    def on_toggle_state_set(self, toggle, state):
        self.confman.conf[self.conf_key] = state
        self.confman.save_conf()
        if self.signal:
            self.confman.emit(self.signal, '')


class PreferencesEntryRow(Handy.ActionRow):
    """
    A preferences row with a title and a button
    title: the title shown
    conf_key: the key of the configuration dictionary/json in ConfManager
    onchange: an optional function that will be called when the entry changes
    subtitle: an optional subtitle to be shown
    signal: an optional signal to let ConfManager emit when the entry changes
    """
    def __init__(self, title, conf_key, onchange=None, subtitle=None,
                 signal=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.title = title
        self.conf_key = conf_key
        self.subtitle = subtitle
        if self.subtitle is not None:
            self.set_subtitle(self.subtitle)
        self.confman = ConfManager()
        self.set_title(self.title)
        self.signal = signal
        self.onchange = onchange

        self.entry = Gtk.Entry()
        self.entry.set_text(self.confman.conf[self.conf_key])
        self.entry.connect('changed', self.on_entry_changed)
        self.add(self.entry)

    def on_entry_changed(self, *args):
        self.confman.conf[self.conf_key] = self.entry.get_text().strip()
        if self.onchange is not None:
            self.onchange(self.confman)
        if self.signal:
            self.confman.emit(self.signal, '')
        self.confman.save_conf()


class GeneralPreferencesPage(Handy.PreferencesPage):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.set_title(_('General'))
        self.set_icon_name('preferences-other-symbolic')

        self.general_preferences_group = Handy.PreferencesGroup()
        self.general_preferences_group.set_title(_('General Settings'))
        toggle_settings = [
            {
                'title': _('Refresh articles on startup'),
                'conf_key': 'refresh_on_startup',
                'signal': None
            },
            {
                'title': _('Show newer articles first'),
                'conf_key': 'new_first',
                'signal': 'gfeeds_new_first_changed'
            },
            {
                'title': _('Open links in your browser'),
                'conf_key': 'open_links_externally',
                'signal': None
            },
            {
                'title': _('Open YouTube links via your video player'),
                'conf_key': 'open_youtube_externally',
                'subtitle': _(
                    'Great when paired with youtube-dl enabled players'
                ),
                'signal': None
            }
        ]
        for s in toggle_settings:
            row = PreferencesToggleRow(
                s['title'],
                s['conf_key'],
                s['signal'],
                subtitle=s['subtitle'] if 'subtitle' in s.keys() else None
            )
            self.general_preferences_group.add(row)
        self.general_preferences_group.add(PreferencesEntryRow(
            _('Preferred media player'), 'media_player'
        ))
        self.general_preferences_group.add(
            PreferencesSpinButtonRow(
                _('Maximum article age'),
                1,
                9999,
                'max_article_age_days',
                subtitle=_('In days')
            )
        )
        self.add(self.general_preferences_group)

        self.cache_preferences_group = Handy.PreferencesGroup()
        self.cache_preferences_group.set_title(_('Cache'))
        button_settings = [
            {
                'title': _('Clear all caches'),
                'button_label': _('Clear caches'),
                'onclick': self.clear_caches,
                'button_style_class': 'destructive-action',
                'signal': 'gfeeds_repopulation_required'
            }
        ]
        for s in button_settings:
            row = PreferencesButtonRow(
                s['title'],
                s['button_label'],
                s['onclick'],
                s['button_style_class'],
                s['signal']
            )
            self.cache_preferences_group.add(row)
        self.add(self.cache_preferences_group)

        self.show_all()

    def clear_caches(self, confman, *args):
        for p in [
                confman.cache_path,
                confman.thumbs_cache_path,
                confman.saved_cache_path
        ]:
            files = [
                abspath(join(p, f)) for f in listdir(p)
            ]
            for f in files:
                if isfile(f):
                    remove(f)
        confman.conf['saved_items'] = {}
        confman.save_conf()


class ViewPreferencesPage(Handy.PreferencesPage):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.set_title(_('View'))
        self.set_icon_name('applications-graphics-symbolic')

        self.view_preferences_group = Handy.PreferencesGroup()
        self.view_preferences_group.set_title(_('View Settings'))
        toggle_settings = [
            {
                'title': _('Show full articles titles'),
                'conf_key': 'full_article_title',
                'signal': 'gfeeds_full_article_title_changed'
            },
            {
                'title': _('Show full feeds names'),
                'conf_key': 'full_feed_name',
                'signal': 'gfeeds_full_feed_name_changed'
            },
            {
                'title': _('Use dark theme for reader mode'),
                'conf_key': 'dark_reader',
                'signal': None
            },
            {
                'title': _('Enable JavaScript'),
                'conf_key': 'enable_js',
                'signal': 'gfeeds_webview_settings_changed'
            },
        ]
        for s in toggle_settings:
            row = PreferencesToggleRow(s['title'], s['conf_key'], s['signal'])
            self.view_preferences_group.add(row)
        self.add(self.view_preferences_group)

        self.show_all()


class AdvancedPreferencesPage(Handy.PreferencesPage):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.set_title(_('Advanced'))
        self.set_icon_name('system-run-symbolic')

        self.advanced_preferences_group = Handy.PreferencesGroup()
        self.advanced_preferences_group.set_title(_('Advanced Settings'))
        spinbtn_settings = [
            {
                'title': _('Maximum refresh threads'),
                'min': 1,
                'max': 32,
                'conf_key': 'max_refresh_threads',
                'signal': None,
                'subtitle': _('How many threads to refresh feeds')
            },
        ]
        for s in spinbtn_settings:
            row = PreferencesSpinButtonRow(
                s['title'],
                s['min'],
                s['max'],
                s['conf_key'],
                s['signal'],
                s['subtitle']
            )
            self.advanced_preferences_group.add(row)
        self.add(self.advanced_preferences_group)

        self.show_all()


class GFeedsSettingsWindow(Handy.PreferencesWindow):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.confman = ConfManager()

        self.pages = [
            GeneralPreferencesPage(),
            ViewPreferencesPage(),
            AdvancedPreferencesPage()
        ]
        for p in self.pages:
            self.add(p)
        self.set_default_size(360, 600)
        # Unneded for Handy 1?
        # self.get_titlebar().set_show_close_button(True)

        self.accel_group = Gtk.AccelGroup()
        self.accel_group.connect(
            *Gtk.accelerator_parse('Escape'), Gtk.AccelFlags.VISIBLE,
            lambda *args: self.close()
        )
        self.add_accel_group(self.accel_group)



import threading
import requests
from gi.repository import Gtk


def __is_online_async_worker(ret_value_l):
    TEST_URL = 'http://httpbin.org/robots.txt'
    EXPECTED_STATUS = 405
    try:
        res = requests.post(TEST_URL)
        ret_value_l[0] = res.status_code == EXPECTED_STATUS
    except requests.exceptions.ConnectionError:
        ret_value_l[0] = False


def is_online():
    t_ret = [False]
    t = threading.Thread(
        group=None,
        target=__is_online_async_worker,
        name=None,
        args=(t_ret,)
    )
    t.start()
    while t.is_alive():
        while Gtk.events_pending():
            Gtk.main_iteration()
    return t_ret[0]



from gi.repository import Gtk, Handy


class GFeedsSearchbar(Handy.SearchBar):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.set_hexpand(False)
        self.entry = Gtk.Entry()
        self.entry.set_icon_from_icon_name(
            Gtk.EntryIconPosition.PRIMARY,
            'system-search-symbolic'
        )
        self.add(self.entry)
        self.set_show_close_button(False)
        self.set_search_mode(False)
        self.connect_entry(self.entry)
        self.show_all()
        # self.set_size_request(360, -1)



from xml.sax.saxutils import escape
from gettext import gettext as _
from gi.repository import Gtk
from gfeeds.opml_manager import opml_to_rss_list
from gfeeds.scrolled_message_dialog import ScrolledMessageDialog


class GFeedsConfirmAddDialog(ScrolledMessageDialog):
    def __init__(self, parent, f_path, http=False, **kwargs):
        super().__init__(
            parent,
            Gtk.DialogFlags.MODAL | Gtk.DialogFlags.DESTROY_WITH_PARENT,
            Gtk.MessageType.QUESTION,
            Gtk.ButtonsType.YES_NO,
            (
                _('Do you want to import these feeds?') if not http
                else _('Do you want to import this feed?')
            ),
            **kwargs
        )

        if not http:
            self.format_secondary_markup(
                escape([f['feed'] for f in opml_to_rss_list(f_path)])
            )
        else:
            self.format_secondary_markup(
                escape(f_path)
            )



from gfeeds.rss_parser import parse_feed
from gfeeds.download_manager import extract_feed_url_from_html
from typing import Optional
from os.path import isfile


def get_feed_link_from_str(feed_str: str) -> Optional[str]:
    fp_feed = parse_feed(feed_str)
    if fp_feed is None:
        return None
    self_link = next(
        (link.href for link in fp_feed.feed.links if link.rel == 'self'),
        None
    )
    if self_link is not None and self_link[:4].lower() == 'http':
        return self_link
    entries = fp_feed.get('entries', [])
    if len(entries) <= 0:
        return None
    fp_item_link = entries[0].get('link', None)
    if fp_item_link is None:
        return None
    return extract_feed_url_from_html(fp_item_link)


def get_feed_link_from_file(feed_path: str) -> Optional[str]:
    if not isfile(feed_path):
        return None
    with open(feed_path, 'r') as fd:
        return get_feed_link_from_str(fd.read())



from datetime import datetime
import pytz


# still not 100% perfect, but good enough, I got tired
def get_date_format(dt):
    now = datetime.now(pytz.utc)
    dtz = dt + (datetime.now()-datetime.utcnow())
    delta = now - dtz + (datetime.now()-datetime.utcnow())
    if dtz.year == now.year and dtz.month == now.month and dtz.day == now.day:
        return '%X'
    elif delta.total_seconds() < 6*24*60*60:
        return '%a %X'
    elif delta.total_seconds() < 365*60*60:
        return '%-e %b %X'
    else:
        return '%-e %b %Y %X'



from gettext import gettext as _
from gi.repository import Gtk, Gdk, GObject, Handy
from xml.sax.saxutils import escape
from gfeeds.confManager import ConfManager
from gfeeds.feeds_manager import FeedsManager
from gfeeds.feeds_view import (
    FeedsViewListbox,
    FeedsViewListboxRow
)
from gfeeds.scrolled_message_dialog import ScrolledMessageDialog
from functools import reduce
from operator import or_


class ManageTagsListboxRow(Gtk.ListBoxRow):
    __gsignals__ = {
        'tag_deleted': (
            GObject.SignalFlags.RUN_FIRST,
            None,
            (str,)  # tag deleted
        ),
    }

    def __init__(self, tag, active=True, **kwargs):
        super().__init__(**kwargs)
        self.builder = Gtk.Builder.new_from_resource(
            '/org/gabmus/gfeeds/ui/manage_tags_listbox_row_content.glade'
        )
        self.tag = tag
        self.main_box = self.builder.get_object('main_box')
        self.label = self.builder.get_object('label')
        self.checkbox = self.builder.get_object('checkbox')
        self.delete_btn = self.builder.get_object('delete_btn')
        self.checkbox.set_active(active)
        self.label.set_text(self.tag)

        self.checkbox_handler_id = self.checkbox.connect(
            'toggled',
            self.on_checkbox_toggled
        )

        self.delete_btn.connect(
            'clicked',
            lambda *args: self.emit('tag_deleted', self.tag)
        )

        self.add(self.main_box)

    def on_checkbox_toggled(self, checkbox):
        with checkbox.handler_block(self.checkbox_handler_id):
            checkbox.set_inconsistent(False)
            checkbox.set_active(not checkbox.get_active())
        self.emit('activate')


class ManageTagsPopover(Gtk.Popover):
    __gsignals__ = {
        'new_tag_added': (
            GObject.SignalFlags.RUN_FIRST,
            None,
            (str,)  # tag added
        ),
        # removed or deleted?
        # removed: removed from selected feeds
        # deleted: deleted from the whole app
        'tag_removed': (
            GObject.SignalFlags.RUN_FIRST,
            None,
            (str,)  # tag removed
        ),
        'tag_deleted': (
            GObject.SignalFlags.RUN_FIRST,
            None,
            (str,)  # tag deleted
        ),
    }

    def __init__(self, relative_to, window, **kwargs):
        super().__init__(**kwargs)
        self.builder = Gtk.Builder.new_from_resource(
            '/org/gabmus/gfeeds/ui/manage_tags_popover_content.glade'
        )
        self.confman = ConfManager()
        self.relative_to = relative_to
        self.window = window
        self.set_relative_to(self.relative_to)
        self.set_modal(True)
        self.main_box = self.builder.get_object('main_box')
        self.add_tag_btn = self.builder.get_object('add_tag_btn')
        self.tags_entry = self.builder.get_object('tags_entry')
        self.tags_listbox = self.builder.get_object('tags_listbox')

        self.tags_listbox.connect(
            'row-activated',
            self.on_tags_listbox_row_activated
        )

        self.tags_listbox.set_sort_func(
            self.tags_listbox_sorting_func,
            None,
            False
        )

        self.tags_entry.connect('changed', self.on_tags_entry_changed)
        self.tags_entry.connect(
            'activate',
            lambda *args: self.on_add_new_tag(
                self.tags_entry.get_text().strip()
            )
        )
        self.add_tag_btn.connect(
            'clicked',
            lambda *args: self.on_add_new_tag(
                self.tags_entry.get_text().strip()
            )
        )

        self.add(self.main_box)
        self.populate_listbox()

    def tags_listbox_sorting_func(self, row1, row2, data, notify_destroy):
        return row1.tag.lower() > row2.tag.lower()

    def populate_listbox(self):
        while True:
            row = self.tags_listbox.get_row_at_index(0)
            if row:
                self.remove(row)
            else:
                break
        for tag in self.confman.conf['tags']:
            self.tags_listbox_add_row(tag, False)
        self.tags_listbox.show_all()

    def on_tags_listbox_row_activated(self, listbox, row):
        with row.checkbox.handler_block(row.checkbox_handler_id):
            row.checkbox.set_inconsistent(False)
            row.checkbox.set_active(not row.checkbox.get_active())
        if row.checkbox.get_active():
            self.on_add_new_tag(row.tag)
        else:
            self.emit('tag_removed', row.tag)

    def tags_listbox_add_row(self, tag: str, show_all=True):
        n_row = ManageTagsListboxRow(tag)
        self.tags_listbox.add(n_row)
        n_row.connect('tag_deleted', self.on_tag_deleted)
        if show_all:
            self.tags_listbox.show_all()

    def tags_listbox_get_row_by_tag(self, tag):
        for row in self.tags_listbox.get_children():
            if row.tag == tag:
                return row
        return None

    def on_add_new_tag(self, n_tag):
        if n_tag == '':
            return
        self.tags_entry.set_text('')
        self.emit('new_tag_added', n_tag)
        if len(self.tags_listbox.get_children()) <= 0 or not reduce(or_, [
                row.tag.lower() == n_tag.lower()
                for row in self.tags_listbox.get_children()
        ]):
            self.tags_listbox_add_row(n_tag)

    def on_tags_entry_changed(self, *args):
        self.add_tag_btn.set_sensitive(
            self.tags_entry.get_text().strip() != ''
        )

    def on_tag_deleted(self, caller, tag):
        self.tags_listbox.remove(caller)
        self.emit('tag_deleted', tag)

    def popup(self, *args):
        self.add_tag_btn.set_sensitive(False)
        self.tags_entry.set_text('')
        selected_feeds = [f.rss_link for f in self.window.get_selected_feeds()]
        for tag in self.confman.conf['tags']:
            all_have_tag = True
            some_have_tag = False
            for feed in selected_feeds:
                if (
                        'tags' in self.confman.conf['feeds'][feed].keys() and
                        tag in self.confman.conf['feeds'][feed]['tags']
                ):
                    some_have_tag = True
                else:
                    all_have_tag = False
            t_row = self.tags_listbox_get_row_by_tag(tag)
            if t_row is not None:
                with t_row.checkbox.handler_block(t_row.checkbox_handler_id):
                    t_row.checkbox.set_inconsistent(False)
                    if some_have_tag and not all_have_tag:
                        t_row.checkbox.set_inconsistent(True)
                    elif all_have_tag:
                        t_row.checkbox.set_active(True)
                    else:
                        t_row.checkbox.set_active(False)
        super().popup()


class ManageFeedsHeaderbar(Gtk.HeaderBar):
    def __init__(self, window, **kwargs):
        super().__init__(**kwargs)
        self.confman = ConfManager()

        self.set_title(_('Manage Feeds'))
        self.set_show_close_button(True)

        self.select_all_btn = Gtk.Button.new_from_icon_name(
            'edit-select-all-symbolic',
            Gtk.IconSize.BUTTON
        )
        self.select_all_btn.set_tooltip_text(_('Select/Unselect all'))

        self.delete_btn = Gtk.Button.new_from_icon_name(
            'user-trash-symbolic',
            Gtk.IconSize.BUTTON
        )
        self.delete_btn.set_tooltip_text(_('Delete selected feeds'))
        self.delete_btn.get_style_context().add_class('destructive-action')

        self.tags_btn = Gtk.Button.new_from_icon_name(
            'tag-symbolic',
            Gtk.IconSize.BUTTON
        )
        self.tags_btn.set_tooltip_text(_('Manage tags for selected feeds'))

        self.pack_end(self.delete_btn)
        self.pack_start(self.select_all_btn)
        self.pack_start(self.tags_btn)

        self.set_actions_sensitive(False)

        self.manage_tags_popover = ManageTagsPopover(
            self.tags_btn,
            window
        )
        self.tags_btn.connect(
            'clicked',
            lambda *args: self.manage_tags_popover.popup()
        )

    def set_actions_sensitive(self, state):
        for w in (self.delete_btn, self.tags_btn):
            w.set_sensitive(state)


class ManageFeedsListboxRow(FeedsViewListboxRow):
    def __init__(self, feed, **kwargs):
        super().__init__(feed, **kwargs)
        self.checkbox.set_no_show_all(False)
        self.checkbox_handler_id = self.checkbox.connect(
            'toggled',
            self.on_checkbox_toggled
        )

    def on_checkbox_toggled(self, checkbox):
        with checkbox.handler_block(self.checkbox_handler_id):
            checkbox.set_active(not checkbox.get_active())
        self.emit('activate')


class ManageFeedsListbox(FeedsViewListbox):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.set_selection_mode(Gtk.SelectionMode.NONE)

    def add_feed(self, feed):
        self.add(ManageFeedsListboxRow(feed))

    def on_row_activated(self, listbox, row):
        with row.checkbox.handler_block(row.checkbox_handler_id):
            row.checkbox.set_active(not row.checkbox.get_active())


class ManageFeedsScrolledWindow(Gtk.ScrolledWindow):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.listbox = ManageFeedsListbox()
        self.set_size_request(360, 500)
        self.set_policy(Gtk.PolicyType.NEVER, Gtk.PolicyType.AUTOMATIC)
        self.add(self.listbox)


class DeleteFeedsConfirmMessageDialog(ScrolledMessageDialog):
    def __init__(self, parent, selected_feeds, **kwargs):
        super().__init__(
            parent,
            Gtk.DialogFlags.MODAL | Gtk.DialogFlags.DESTROY_WITH_PARENT,
            Gtk.MessageType.QUESTION,
            Gtk.ButtonsType.YES_NO,
            _('Do you want to delete these feeds?'),
            **kwargs
        )

        self.format_secondary_markup(
            '\n'.join([escape(f.title) for f in selected_feeds])
        )


class GFeedsManageFeedsWindow(Handy.Window):
    def __init__(self, appwindow, **kwargs):
        super().__init__(**kwargs)
        self.appwindow = appwindow
        self.confman = ConfManager()
        self.feedman = FeedsManager()
        self.main_box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
        self.scrolled_window = ManageFeedsScrolledWindow()
        self.listbox = self.scrolled_window.listbox
        self.headerbar = ManageFeedsHeaderbar(self)

        self.headerbar.delete_btn.connect(
            'clicked',
            self.on_delete_clicked
        )
        self.headerbar.select_all_btn.connect(
            'clicked',
            self.on_select_all_clicked
        )
        self.listbox.connect('row-activated', self.on_row_activated)
        self.set_title(_('Manage Feeds'))

        self.window_handle = Handy.WindowHandle()
        self.window_handle.add(self.headerbar)
        self.main_box.pack_start(self.window_handle, False, False, 0)
        self.window_handle.set_vexpand(False)

        self.main_box.pack_start(self.scrolled_window, True, True, 0)
        self.add(self.main_box)

        self.accel_group = Gtk.AccelGroup()
        self.accel_group.connect(
            *Gtk.accelerator_parse('Escape'), Gtk.AccelFlags.VISIBLE,
            lambda *args: self.close()
        )
        self.add_accel_group(self.accel_group)

        self.set_type_hint(Gdk.WindowTypeHint.DIALOG)
        self.set_modal(True)
        self.set_transient_for(self.appwindow)

        self.headerbar.manage_tags_popover.connect(
            'new_tag_added',
            self.on_new_tag_added
        )
        self.headerbar.manage_tags_popover.connect(
            'tag_removed',
            self.on_tag_removed
        )
        self.headerbar.manage_tags_popover.connect(
            'tag_deleted',
            self.on_tag_deleted
        )

    def on_new_tag_added(self, caller, tag):
        self.confman.add_tag(
            tag,
            [feed.rss_link for feed in self.get_selected_feeds()]
        )

    def on_tag_removed(self, caller, tag):
        self.confman.remove_tag(
            tag,
            [feed.rss_link for feed in self.get_selected_feeds()]
        )

    def on_tag_deleted(self, caller, tag):
        self.confman.delete_tag(tag)

    def present(self, *args, **kwargs):
        super().present(*args, **kwargs)
        self.show_all()

    def get_selected_feeds(self):
        return [
            row.feed
            for row in self.listbox.get_children()
            if row.checkbox.get_active()
        ]

    def on_delete_clicked(self, *args):
        selected_feeds = self.get_selected_feeds()
        dialog = DeleteFeedsConfirmMessageDialog(self, selected_feeds)
        res = dialog.run()
        dialog.close()
        if res == Gtk.ResponseType.YES:
            self.feedman.delete_feeds(selected_feeds)
            self.headerbar.set_actions_sensitive(False)

    def on_select_all_clicked(self, *args):
        unselect = True
        for row in self.listbox.get_children():
            if not row.checkbox.get_active():
                unselect = False
                row.emit('activate')
        if unselect:
            for row in self.listbox.get_children():
                row.emit('activate')

    def on_row_activated(self, listbox, activated_row):
        for row in listbox.get_children():
            if row.checkbox.get_active():
                self.headerbar.set_actions_sensitive(True)
                return
        self.headerbar.set_actions_sensitive(False)



from gi.repository import Gtk, Handy, GObject
from typing import Callable
from gfeeds.confManager import ConfManager
from gfeeds.feeds_manager import FeedsManager
from gfeeds.sidebar import GFeedsSidebar
from gfeeds.headerbar import GFeedHeaderbar
from gfeeds.searchbar import GFeedsSearchbar
from gfeeds.suggestion_bar import (
    GFeedsConnectionBar,
    GFeedsErrorsBar
)
from gfeeds.webview import GFeedsWebView
from gfeeds.stack_with_empty_state import StackWithEmptyState
from functools import reduce
from operator import or_
from subprocess import Popen


class GFeedsAppWindow(Handy.ApplicationWindow):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.confman = ConfManager()
        self.feedman = FeedsManager()

        self.set_title('Feeds')
        self.set_icon_name('org.gabmus.gfeeds')

        self.sidebar = GFeedsSidebar()
        self.sidebar.listbox.connect(
            'row-activated',
            self.on_sidebar_row_activated
        )
        self.sidebar.saved_items_listbox.connect(
            'row-activated',
            self.on_sidebar_row_activated
        )

        self.webview = GFeedsWebView()

        separator = Gtk.Separator()
        separator.get_style_context().add_class('sidebar')

        leaflet_builder = Gtk.Builder.new_from_resource(
            '/org/gabmus/gfeeds/ui/gfeeds_leaflet.glade'
        )
        self.leaflet = leaflet_builder.get_object('leaflet')
        self.connection_bar = GFeedsConnectionBar()
        self.errors_bar = GFeedsErrorsBar(self)
        self.feedman.connect(
            'feedmanager_refresh_end',
            lambda *args: self.errors_bar.engage(
                self.feedman.errors,
                self.feedman.problematic_feeds
            )
        )
        self.searchbar = GFeedsSearchbar()
        self.searchbar.entry.connect(
            'changed',
            lambda entry: self.sidebar.set_search(entry.get_text())
        )
        self.searchbar.connect(
            'notify::search-mode-enabled',
            lambda caller, enabled:
                self.headerbar.search_btn.set_active(caller.get_search_mode())
        )
        self.sidebar_box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
        self.sidebar_box.set_size_request(360, 100)
        self.sidebar_box.pack_start(self.searchbar, False, False, 0)
        self.sidebar_box.pack_start(self.connection_bar, False, False, 0)
        self.sidebar_box.pack_start(self.errors_bar, False, False, 0)
        self.stack_with_empty_state = StackWithEmptyState(self.sidebar)
        self.sidebar_box.pack_start(self.stack_with_empty_state, True, True, 0)
        self.leaflet.add(self.sidebar_box)
        self.leaflet.add(separator)
        self.leaflet.add(self.webview)
        self.leaflet.child_set_property(separator, 'allow-visible', False)
        self.leaflet.connect('notify::folded', self.on_main_leaflet_folded)

        self.swipe_group = Handy.SwipeGroup()

        self.size_group_left = Gtk.SizeGroup(
            mode=Gtk.SizeGroupMode.HORIZONTAL
        )
        self.size_group_right = Gtk.SizeGroup(
            mode=Gtk.SizeGroupMode.HORIZONTAL
        )
        self.size_group_left.add_widget(self.sidebar_box)
        self.size_group_right.add_widget(self.webview)
        self.headerbar = GFeedHeaderbar(
            self.size_group_left,
            self.size_group_right,
            self.on_back_button_clicked,
            self.webview,
            self.searchbar
        )

        self.swipe_group.add_swipeable(self.leaflet)
        self.swipe_group.add_swipeable(self.headerbar.leaflet)

        self.headerbar.stack_switcher.set_stack(self.sidebar)
        self.headerbar.connect(
            'gfeeds_headerbar_squeeze',
            self.on_headerbar_squeeze
        )

        self.main_box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)

        self.bottom_bar = Handy.ViewSwitcherBar()
        self.bottom_bar.set_stack(self.sidebar)
        self.sidebar_box.pack_end(self.bottom_bar, False, False, 0)

        self.set_headerbar()

        # listening on the headerbar leaflet visible-child because of a bug in
        # libhandy that doesn't notify the correct child on the main leaflet
        self.headerbar.leaflet.connect(
            'notify::visible-child', self.on_main_leaflet_folded
        )

        self.add(self.main_box)
        self.main_box.pack_end(self.leaflet, True, True, 0)

        # Why this -52?
        # because every time a new value is saved, for some reason
        # it's the actual value +52 out of nowhere
        # this makes the window ACTUALLY preserve its old size
        self.resize(
            self.confman.conf['windowsize']['width']-52,
            self.confman.conf['windowsize']['height']-52
        )
        self.size_allocation = self.get_allocation()
        self.connect('size-allocate', self.update_size_allocation)
        self.on_main_leaflet_folded()

        # accel_group is for keyboard shortcuts
        self.accel_group = Gtk.AccelGroup()
        self.add_accel_group(self.accel_group)
        shortcuts_l = [
            {
                'combo': 'F10',
                'cb': lambda *args: self.headerbar.menu_btn.clicked()
            },
            {
                'combo': '<Control>r',
                'cb': self.feedman.refresh
            },
            {
                'combo': '<Control>f',
                'cb': lambda *args: self.headerbar.search_btn.set_active(True)
            },
            {
                'combo': '<Control>j',
                'cb': self.sidebar.select_next_article
            },
            {
                'combo': '<Control>k',
                'cb': self.sidebar.select_prev_article
            },
            {
                'combo': '<Control>plus',
                'cb': self.webview.key_zoom_in
            },
            {
                'combo': '<Control>minus',
                'cb': self.webview.key_zoom_out
            },
            {
                'combo': '<Control>equal',
                'cb': self.webview.key_zoom_reset
            }
        ]
        for s in shortcuts_l:
            self.add_accelerator(s['combo'], s['cb'])

    def on_headerbar_squeeze(self, caller: GObject.Object, squeezed: bool):
        self.bottom_bar.set_reveal(squeezed)

    def set_headerbar(self, *args):
        self.main_box.pack_start(self.headerbar, False, False, 0)
        self.headerbar.set_vexpand(False)

    def add_accelerator(self, shortcut: str, callback: Callable):
        if shortcut:
            key, mod = Gtk.accelerator_parse(shortcut)
            self.accel_group.connect(
                key, mod, Gtk.AccelFlags.VISIBLE, callback
            )

    def emit_destroy(self, *args):
        self.emit('destroy')

    def on_destroy(self, *args):
        self.confman.conf['windowsize'] = {
            'width': self.size_allocation.width,
            'height': self.size_allocation.height
        }
        # cleanup old read items
        feeds_items_links = [fi.link for fi in self.feedman.feeds_items]
        to_rm = []
        for ri in self.confman.conf['read_items']:
            if ri not in feeds_items_links:
                to_rm.append(ri)
        for ri in to_rm:
            self.confman.conf['read_items'].remove(ri)
        self.confman.save_conf()

    def on_sidebar_row_activated(
            self,
            listbox: Gtk.ListBox,
            row: Gtk.ListBoxRow
    ):
        row.popover.set_read(True)
        other_listbox = (
            self.sidebar.listbox
            if listbox == self.sidebar.saved_items_listbox
            else self.sidebar.saved_items_listbox
        )
        for other_row in other_listbox:
            if other_row.feeditem.link == row.feeditem.link:
                other_row.popover.set_read(True)
                break
        if (
                self.confman.conf['open_youtube_externally'] and
                reduce(or_, [
                    f'://{pfx}' in row.feeditem.link
                    for pfx in [
                        p + 'youtube.com'
                        for p in ('', 'www.', 'm.')
                    ]
                ])
        ):
            cmd_parts = [
                self.confman.conf["media_player"],
                row.feeditem.link
            ]
            if self.confman.is_flatpak:
                cmd_parts.insert(0, 'flatpak-spawn --host')
            cmd = ' '.join(cmd_parts)
            Popen(
                cmd,
                shell=True
            )
            return
        self.webview.load_feeditem(row.feeditem)
        self.headerbar.set_article_title(
            row.feeditem.title
        )
        self.headerbar.share_btn.set_sensitive(True)
        self.headerbar.open_externally_btn.set_sensitive(True)
        self.leaflet.set_visible_child(self.webview)
        self.headerbar.leaflet.set_visible_child(
            self.headerbar.right_headerbar
        )
        self.on_main_leaflet_folded()
        listbox.invalidate_filter()
        other_listbox.invalidate_filter()

    def on_main_leaflet_folded(self, *args):
        # target = None
        # other = None
        if self.leaflet.get_folded():
            # target = self.headerbar.leaflet.get_visible_child()
            self.headerbar.back_button.show()
            self.headerbar.stack_switcher.set_no_show_all(False)
            self.headerbar.stack_switcher.show()
            self.headerbar.squeezer.set_child_enabled(
                self.headerbar.stack_switcher, True
            )
        else:
            self.headerbar.back_button.hide()
            self.headerbar.stack_switcher.set_no_show_all(True)
            self.headerbar.stack_switcher.hide()
            self.headerbar.squeezer.set_child_enabled(
                self.headerbar.stack_switcher, False
            )
        # self.headerbar.headergroup.set_focus(target)

    def on_back_button_clicked(self, *args):
        self.leaflet.set_visible_child(self.sidebar_box)
        self.on_main_leaflet_folded()
        self.sidebar.listbox.select_row(None)
        self.sidebar.saved_items_listbox.select_row(None)

    def update_size_allocation(self, *args):
        self.size_allocation = self.get_allocation()
